#    Strome - Micro-framework for Python for building processing pipelines
#    Copyright (C) 2021 Dmitry Berezovsky
#    The MIT License (MIT)
#
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
import json
import logging
import os
from abc import ABC
from typing import Dict, Optional, Any, Type

import cli_rack.serialize

LOGGER = logging.getLogger("cache")


class CacheSerializer(ABC):
    META_UPDATED_AT = "updated_at"
    META_PERSISTED_AT = "persisted_at"
    META_LOADED_AT = "loaded_at"
    META_NAME = "cache_name"
    CACHE_DATA = "data"

    def persist(self, cache: "Cache", data: dict):
        raise NotImplementedError

    def load(self) -> Dict:
        raise NotImplementedError


class InMemoryCacheSerializer(CacheSerializer):
    def persist(self, cache: "Cache", data: dict):
        pass

    def load(self) -> Dict:
        return {}


class FileSystemCacheSerializer(CacheSerializer):
    def __init__(self, file_path: str) -> None:
        super().__init__()
        self.file_path = file_path

    def persist(self, cache: "Cache", data: dict):
        with open(self.file_path, "w") as f:
            cache_dict = {
                CacheSerializer.META_UPDATED_AT: cache.last_updated_at,
                CacheSerializer.META_PERSISTED_AT: cache.persisted_at,
                CacheSerializer.META_NAME: cache.name,
                CacheSerializer.CACHE_DATA: data,
            }
            json.dump(cache_dict, f, cls=cli_rack.serialize.DateTimeEncoder)

    def load(self) -> Dict:
        if os.path.isfile(self.file_path):
            with open(self.file_path, "r") as f:
                try:
                    return json.load(f, cls=cli_rack.serialize.DateTimeDecoder)
                except json.decoder.JSONDecodeError as e:
                    LOGGER.warning(
                        "Unable to read cache file {} (REASON: {}). Cache might be corrupted. "
                        "Starting with empty cache. ".format(self.file_path, str(e))
                    )

        return {}


class Cache(object):
    def __init__(self, name="default", serializer=InMemoryCacheSerializer()) -> None:
        self.__last_updated: Optional[datetime.datetime] = None
        self.__persisted_at: Optional[datetime.datetime] = None
        self.__loaded_at: Optional[datetime.datetime] = None
        self.__cache: Dict[str, Any] = {}
        self.__serializer = serializer
        self.name: str = name

    def get(self, key: str, default_val: Any = None) -> Any:
        return self.__cache.get(key, default_val)

    def put(self, key: str, val: Any) -> None:
        self.__cache[key] = val
        self.__last_updated = datetime.datetime.now()

    def has(self, key: str):
        return key in self.__cache

    def get_or_raise(self, key: str, exc: Type[Exception] = KeyError):
        if self.has(key):
            return self.get(key)
        else:
            raise exc('Cache {} doesn\'t have key "{}"'.format(self.name, key))

    def load(self) -> None:
        data_dict = self.__serializer.load()
        self.__loaded_at = datetime.datetime.now()
        self.__persisted_at = data_dict.get(CacheSerializer.META_PERSISTED_AT)
        self.__last_updated = data_dict.get(CacheSerializer.META_UPDATED_AT)
        self.__cache = data_dict.get(CacheSerializer.CACHE_DATA, {})

    def persist(self) -> None:
        self.__serializer.persist(self, self.__cache)

    @property
    def has_changes(self):
        return self.__last_updated is not None

    @property
    def persisted_at(self) -> Optional[datetime.datetime]:
        self.__persisted_at = datetime.datetime.now()
        return self.__persisted_at

    @property
    def last_updated_at(self) -> Optional[datetime.datetime]:
        return self.__last_updated

    @property
    def loaded_at(self) -> Optional[datetime.datetime]:
        return self.__loaded_at



#    Strome - Micro-framework for Python for building processing pipelines
#    Copyright (C) 2021 Dmitry Berezovsky
#    The MIT License (MIT)
#
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
import os
from os import path
from tempfile import gettempdir
from typing import List, Optional, Dict, Any

import jinja2
from jinja2 import FileSystemLoader

from cli_rack.utils import safe_cast, none_throws
from strome.const import CONF_INPUT_FILES, CONF_CONTEXT, CONF_CONTEXT_FILES, RUNTIME_OUTPUT
from strome.pipeline import PipelineElement, StromeRuntime
from strome.utils import yaml_file_to_dict, deepmerge_dict

__all__ = ("Jinja2Processor",)

PARAM_OUTPUT_DIR = "output-dir"
PARAM_TAG_LIBS = "tag-libs"
PARAM_SEARCH_PATH = "search-path"

RUNTIME_CONTEXT_VARS = "context-vars"
RUNTIME_JINJA2_OUTPUT_FILES = "jinja2-output-files"


class Jinja2Processor(PipelineElement):
    NAME = "jinja2"
    PARAMS_SCHEMA = {
        PARAM_TAG_LIBS: {"type": "list", "schema": {"type": "string"}},
        PARAM_OUTPUT_DIR: {"type": "string"},
        PARAM_SEARCH_PATH: {"type": "list", "schema": {"type": "string"}, "default": [""]},
    }

    def __init__(self) -> None:
        super().__init__()
        self.output_dir = gettempdir()
        self.search_path: List[str] = [""]
        self.jinja_options: Dict[str, Any] = {}
        self.__jinja_env: Optional[jinja2.Environment] = None

    def setup(self, flow_runtime: StromeRuntime, _params: dict):
        # HANDLE PARAMS
        default_output_dir = path.join(flow_runtime.temp_dir, "jinja2")
        params: Dict[str, Any] = none_throws(self._validate_and_normalize_params(_params).normalized_data)
        # PARAM: output dir
        self.output_dir = params.get(PARAM_OUTPUT_DIR, default_output_dir)
        self._ensure_dir(self.output_dir)
        flow_runtime.register_resource_path(self.output_dir)
        # PARAM: Search Path
        self.search_path = safe_cast(List[str], params.get(PARAM_SEARCH_PATH))
        self.search_path.append(os.path.dirname(flow_runtime.config_path))

        # Handle context files
        if CONF_CONTEXT not in flow_runtime.flow_config:
            flow_runtime.flow_config[CONF_CONTEXT] = {}
        context_vars = flow_runtime.flow_config[CONF_CONTEXT]
        if not isinstance(context_vars, dict):
            raise ValueError("{} must be dictionary".format(CONF_CONTEXT))

        # Loading context files
        if CONF_CONTEXT_FILES in flow_runtime.flow_config:
            if not isinstance(flow_runtime.flow_config[CONF_CONTEXT_FILES], list):
                raise ValueError(
                    "{} must be a list".format(CONF_CONTEXT_FILES),
                )
            included_context: Dict[str, Any] = {}
            for context_file in flow_runtime.flow_config[CONF_CONTEXT_FILES]:
                ctx = yaml_file_to_dict(context_file)
                included_context = deepmerge_dict(included_context, ctx)
            context_vars = deepmerge_dict(included_context, context_vars)
            flow_runtime.context[RUNTIME_CONTEXT_VARS] = context_vars

        # Setup Jinja2
        self.__jinja_env = jinja2.Environment(
            comment_start_string="{##",
            comment_end_string="##}",
            loader=FileSystemLoader(searchpath=self.search_path, followlinks=True),
            **self.jinja_options,
        )

    def prepare_rendering_context(self, context: dict, flow_runtime: StromeRuntime):
        pass

    def run(self, flow_runtime: StromeRuntime):
        rendering_context = dict(
            config=flow_runtime.config,
            output=flow_runtime.context[RUNTIME_OUTPUT],
            build_time=datetime.datetime.now(),
        )
        self.prepare_rendering_context(rendering_context, flow_runtime)
        rendering_context.update(flow_runtime.context.get(RUNTIME_CONTEXT_VARS, {}))
        flow_runtime.context[RUNTIME_JINJA2_OUTPUT_FILES] = output_files = []
        for file_path in flow_runtime.flow_config.get(CONF_INPUT_FILES, tuple()):
            file_name = path.basename(file_path)
            output_file_path = path.join(self.output_dir, file_name)
            template = none_throws(self.__jinja_env).get_template(file_path, globals=rendering_context)  # type: ignore
            with open(output_file_path, "w") as f:
                f.write(template.render())
                output_files.append(output_file_path)
                flow_runtime.register_input_file(output_file_path)



#    Strome - Micro-framework for Python for building processing pipelines
#    Copyright (C) 2021 Dmitry Berezovsky
#    The MIT License (MIT)
#
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import logging
import os
import sys
from typing import NamedTuple, List, Union, Type

import cli_rack.loader
import cli_rack.utils
from cli_rack.utils import safe_cast
from strome import processors
from strome import const
from strome.error import InvalidModuleError, ConfigValidationError
from strome.pipeline import PipelineElement, StromeRuntime
from strome.utils import deepmerge_dict

LOGGER = logging.getLogger("strome.core")

DEFAULT_PROCESSORS_CONFIG = {
    const.CONF_PIPELINE: ["jinja2"],
    const.CONF_TEMP_DIR: "generated",
    const.CONF_LIB_DIR: "build-libs",
    const.CONF_CLASSPATH: [],
    const.CONF_LIBS: [],
}


class PipelineItem(NamedTuple):
    pipeline_element: PipelineElement
    pipeline_params: dict


Pipeline = List[PipelineItem]


def setup_pipeline(pipeline_def: List[dict], flow_runtime: StromeRuntime) -> Pipeline:
    pipeline: List[PipelineItem] = []
    for cfg in pipeline_def:
        processor_name = cfg if isinstance(cfg, str) else safe_cast(str, cfg.get(const.CONF_PIPELINE_NAME))
        params = {} if isinstance(cfg, str) else cfg.get(const.CONF_PIPELINE_PARAMS, {})
        try:
            if not processors.REGISTRY.exists(processor_name):
                if not __is_fully_qualified_name(processor_name):
                    raise InvalidModuleError("Processor {} is unknown".format(processor_name))
                else:
                    load_module(processor_name)
            pipeline_element = processors.REGISTRY.get_by_name(processor_name)()  # type: ignore
        except KeyError:
            raise ValueError("Invalid pipeline configuration. Processor {} is unknown".format(processor_name))
        pipeline_element.load()
        pipeline.append(PipelineItem(pipeline_element, params))
    for p in pipeline:
        if isinstance(p.pipeline_element.REQUIRES, list):
            for x in p.pipeline_element.REQUIRES:
                if not flow_runtime.has_output(x):
                    raise ValueError(
                        "{} expects {} to be provided by other processor, "
                        "but it is not registered in the context. "
                        "Check if your pipeline is configured correctly".format(p.pipeline_element.name(), x)
                    )
        p.pipeline_element.setup(flow_runtime, p.pipeline_params)
        if isinstance(p.pipeline_element.PROVIDES, list):
            for x in p.pipeline_element.PROVIDES:
                flow_runtime.register_output(p.pipeline_element, x, None)
    return pipeline


def run_pipeline(pipeline: Pipeline, flow_runtime: StromeRuntime):
    for p in pipeline:
        p.pipeline_element.run(flow_runtime)


def __lib_path_resolver(meta: cli_rack.loader.LoadedDataMeta):
    if os.path.isdir(os.path.join(meta.path, "src")):
        return "src"
    return ""


def __is_fully_qualified_name(name: str):
    return "." in name


def load_libs(flow_runtime: StromeRuntime, force_reload=False):
    locators = safe_cast(List, flow_runtime.flow_config.get(const.CONF_LIBS))
    if len(locators) > 0:
        cli_rack.utils.ensure_dir(cli_rack.loader.LoaderRegistry.target_dir)
    for locator in locators:
        if not isinstance(locator, cli_rack.loader.BaseLocatorDef):
            pass
        meta = cli_rack.loader.LoaderRegistry.load(locator, __lib_path_resolver, force_reload)
        flow_runtime.register_lib_meta(meta)


def load_context_path(flow_runtime: StromeRuntime):
    context_path = flow_runtime.flow_config.get(const.CONF_CLASSPATH, [])
    for p in context_path:
        sys.path.append(p)
    for meta in flow_runtime.libs_meta:
        sys.path.append(os.path.join(meta.path, meta.target_path))


def preload_modules(flow_runtime: StromeRuntime):
    modules_to_load = safe_cast(List, flow_runtime.flow_config.get(const.CONF_PRELOAD))
    for m in modules_to_load:
        load_module(m)


def import_class(class_name_or_class: Union[str, Type]):
    if isinstance(class_name_or_class, str):
        try:
            module_name, class_name = class_name_or_class.rsplit(".", 1)
            module = __import__(module_name, globals(), locals(), [class_name], 0)
            clazz = getattr(module, class_name)
        except ImportError as e:
            raise InvalidModuleError("Class doesn't exist: " + str(class_name_or_class), e)
        except AttributeError as e:
            raise InvalidModuleError("Class doesn't exist: " + str(class_name_or_class), e)
    else:
        clazz = class_name_or_class
    # Validations
    assert clazz is not None, "module instance shouldn't be None"
    return clazz


def __import_processor(class_name_or_class: str) -> Type[PipelineElement]:
    clazz = import_class(class_name_or_class)
    if not issubclass(clazz, PipelineElement):
        raise InvalidModuleError("Processor must implement strome.preprocessor.PipelineElement")
    return clazz


def load_module(class_name: str):
    LOGGER.info("Loading module " + class_name)
    processor_cls = __import_processor(class_name)
    processors.REGISTRY.register(processor_cls, aliases=(class_name,))
    LOGGER.info("\t\tModule {}({}) loaded".format(class_name, processor_cls.name()))


def process_strome_config(  # noqa: C901
    config_dict: dict, flow_runtime: StromeRuntime, default_strome_conf: dict = None
):
    if flow_runtime.root_config_element in config_dict:
        flow_runtime.is_configured = True
        # Ensure preprocessor context is declared
        conf_strome = config_dict[flow_runtime.root_config_element]
        if not isinstance(conf_strome, dict):
            raise ValueError("{} must be dictionary".format(flow_runtime.root_config_element))
        if default_strome_conf is None:
            default_strome_conf = DEFAULT_PROCESSORS_CONFIG
        config_dict[flow_runtime.root_config_element] = deepmerge_dict(
            default_strome_conf, conf_strome, merge_lists=False
        )
        # Validate classpath
        classpath = flow_runtime.flow_config.get(const.CONF_CLASSPATH)
        if classpath:
            for path in classpath:
                if isinstance(path, str):
                    if not os.path.exists(path):
                        raise ConfigValidationError(
                            "Path " + path + " doesn't exist",
                            config_section=flow_runtime.root_config_element + "/" + const.CONF_CLASSPATH,
                        )
                else:
                    raise ConfigValidationError(
                        "Context path must be the list of strings",
                        config_section=flow_runtime.root_config_element + "/" + const.CONF_CLASSPATH,
                    )
        # Validate libs
        cli_rack.loader.LoaderRegistry.target_dir = flow_runtime.flow_config.get(const.CONF_LIB_DIR)
        lib_locators: List[cli_rack.loader.BaseLocatorDef] = []
        libs = flow_runtime.flow_config.get(const.CONF_LIBS)
        if libs is not None:
            for i, lib_def in enumerate(libs):
                try:
                    lib_locators.append(cli_rack.loader.LoaderRegistry.parse_locator(lib_def))
                except cli_rack.loader.LoaderError as e:
                    raise ConfigValidationError(
                        "Invalid library definition: " + str(e),
                        config_section=flow_runtime.root_config_element + "/" + const.CONF_LIBS,
                    )
            if len(lib_locators) > 0:
                flow_runtime.flow_config[const.CONF_LIBS] = lib_locators



#    Strome - Micro-framework for Python for building processing pipelines
#    Copyright (C) 2021 Dmitry Berezovsky
#    The MIT License (MIT)
#
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import logging
import os
import subprocess
from copy import copy
from tempfile import gettempdir
from typing import Optional, List, Any, Dict

import cerberus
import cli_rack.loader
import cli_rack.utils

from strome import const
from strome.cache import Cache, FileSystemCacheSerializer
from strome.error import ConfigValidationError


class ValidationResult:
    def __init__(self) -> None:
        self.errors: Dict[str, Any] = {}
        self.normalized_data: Optional[dict] = None
        self.data: Optional[dict] = None

    @property
    def has_errors(self):
        return len(self.errors) > 0


class StromeRuntime:
    def __init__(
        self, config_path: str, config: Optional[dict] = None, root_element_name: str = const.CONF_ROOT_ELEMENT_NAME
    ) -> None:
        self.config_path = config_path
        self.config = config or {}
        self.is_configured = False
        self.__root_element_name = root_element_name
        self.context: Dict[str, Any] = {
            const.RUNTIME_INPUT_FILES: [],
            const.RUNTIME_RESOURCE_PATH: [],
            const.RUNTIME_OUTPUT: {},
        }
        self.__generated_dir = gettempdir()
        self.__libs_meta: List[cli_rack.loader.LoadedDataMeta] = []
        self.is_configured = False

    def init(self):
        if self.flow_config.get(const.CONF_TEMP_DIR) is None:
            self.__generated_dir = os.path.join(self.__generated_dir, "generated")
        else:
            self.__generated_dir = self.flow_config.get(const.CONF_TEMP_DIR)
        os.makedirs(self.__generated_dir, exist_ok=True)
        self.register_resource_path(self.temp_dir)

    @property
    def flow_config(self) -> dict:
        return self.config.get(self.__root_element_name, {})

    @property
    def temp_dir(self) -> str:
        return self.__generated_dir

    @property
    def libs_meta(self) -> List[cli_rack.loader.LoadedDataMeta]:
        return self.__libs_meta

    @property
    def root_config_element(self) -> str:
        return self.__root_element_name

    def register_input_file(self, file_path: str):
        self.context[const.RUNTIME_INPUT_FILES].append(file_path)

    def register_resource_path(self, dir_path: str):
        self.context[const.RUNTIME_RESOURCE_PATH].append(dir_path)

    def register_output(self, pipeline_el: "PipelineElement", key: str, value):
        if key not in pipeline_el.PROVIDES:
            raise ValueError("Attempt to register output data which is not listed in {}.PROVIDES")
        self.context[const.RUNTIME_OUTPUT][key] = value

    def register_lib_meta(self, meta: cli_rack.loader.LoadedDataMeta):
        self.__libs_meta.append(meta)

    def get_output(self, key: str, default_val=None) -> Any:
        return self.context[const.RUNTIME_OUTPUT].get(key, default_val)

    def has_output(self, key: str) -> Any:
        return key in self.context[const.RUNTIME_OUTPUT]


class PipelineElement(object):
    NAME: str
    PARAMS_SCHEMA: dict = {}
    PROVIDES: List[str] = []
    REQUIRES: List[str] = []
    DEPENDENCIES: List[str] = []

    @classmethod
    def name(cls):
        return cls.NAME if cls.NAME is not None else cls.__name__

    def __init__(self) -> None:
        self.logger = logging.getLogger(self.name())
        self.validator = cerberus.Validator()

    def _validate_and_normalize_params(
        self, params, raise_error=True, schema: Optional[dict] = None
    ) -> ValidationResult:
        _schema = schema or self.PARAMS_SCHEMA
        result = ValidationResult()
        result.data = copy(params)
        is_valid = self.validator.validate(params, _schema)
        if not is_valid:
            result.errors = copy(self.validator.errors)
        else:
            result.normalized_data = self.validator.normalized(params, _schema)
        if result.has_errors and raise_error:
            raise ConfigValidationError("Invalid pipeline element configuration", self, result)
        return result

    @staticmethod
    def _ensure_dir(dir_name: str):
        cli_rack.utils.ensure_dir(dir_name)

    def load(self):
        pass

    def setup(self, flow_runtime: StromeRuntime, params: dict):
        pass

    def run(self, flow_runtime: StromeRuntime):
        pass


class ExternalExecutableMixin(object):
    @staticmethod
    def run_executable(*args, hide_output=False, mute_output=False) -> subprocess.CompletedProcess:
        return cli_rack.utils.run_executable(*args, hide_output=hide_output, mute_output=mute_output)

    def is_successful_exit_code(self, *args, expected_code=0) -> bool:
        return self.run_executable(*args, mute_output=True).returncode == expected_code


class CacheMixin(object):
    def __init__(self) -> None:
        self.cache: Cache = Cache()

    def setup_cache(self, pipeline_el: PipelineElement, flow_runtime: StromeRuntime):
        cache_file_location = os.path.join(flow_runtime.temp_dir, "_CACHE", pipeline_el.name() + ".json")
        cli_rack.utils.ensure_dir(os.path.dirname(cache_file_location))
        self.cache = Cache(pipeline_el.name(), serializer=FileSystemCacheSerializer(cache_file_location))
        self.cache.load()
